/*
Copyright 2024.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controller

import (
	"context"
	"fmt"
	"time"

	"gopkg.in/yaml.v2"
	corev1 "k8s.io/api/core/v1"
	k8serrors "k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/utils/ptr"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/log"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"

	diffv1alpha1 "github.com/allanger/gitops-diff-operator/api/v1alpha1"
	k8sbadhouseplantsnetv1alpha1 "github.com/allanger/gitops-diff-operator/api/v1alpha1"
	"github.com/argoproj/argo-cd/v2/pkg/apiclient"
	argocdclient "github.com/argoproj/argo-cd/v2/pkg/apiclient"
	"github.com/argoproj/argo-cd/v2/pkg/apiclient/application"
	argoappv1 "github.com/argoproj/argo-cd/v2/pkg/apis/application/v1alpha1"
	"github.com/sergi/go-diff/diffmatchpatch"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// DiffReconciler reconciles a Diff object
type DiffReconciler struct {
	client.Client
	Scheme *runtime.Scheme
}

// +kubebuilder:rbac:groups=k8s.badhouseplants.net,resources=diffs,verbs=get;list;watch;create;update;patch;delete
// +kubebuilder:rbac:groups=k8s.badhouseplants.net,resources=diffs/status,verbs=get;update;patch
// +kubebuilder:rbac:groups=k8s.badhouseplants.net,resources=diffs/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the Diff object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.18.2/pkg/reconcile
func (r *DiffReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	log := log.FromContext(ctx)
	log.Info("dmmu")

	// TODO: It should not be hardcoded
	reconcilePeriod := 30 * time.Second
	reconcileResult := reconcile.Result{RequeueAfter: reconcilePeriod}

	diff := &diffv1alpha1.Diff{}
	err := r.Client.Get(ctx, types.NamespacedName{
		Namespace: req.Namespace,
		Name:      req.Name,
	}, diff)

	if err != nil {
		if k8serrors.IsNotFound(err) {
			return reconcileResult, nil
		}
		return reconcileResult, err
	}

	source := &diffv1alpha1.AppSource{}
	if err := r.Client.Get(ctx, types.NamespacedName{
		Namespace: req.Namespace,
		Name:      diff.Spec.AppSourceName,
	}, source); err != nil {
		return reconcileResult, err
	}
	client, err := apiclient.NewClient(&argocdclient.ClientOptions{
		ServerAddr: "argocd-server.argo-system.svc.cluster.local:80",
		Insecure:   true,
		AuthToken:  "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJhcmdvY2QiLCJzdWIiOiJkcm9uZTphcGlLZXkiLCJuYmYiOjE3MTk5NTc1OTMsImlhdCI6MTcxOTk1NzU5MywianRpIjoidGVzdC1kaWZmIn0.eJ-MOONUca0JEtoQlfNFdV_U-dULjyg37u8jrc2boYc",
	})
	if err != nil {
		return reconcileResult, err
	}
	closer, appCl, err := client.NewApplicationClient()
	if err != nil {
		return reconcileResult, err
	}
	defer closer.Close()

	qMain := application.ApplicationManifestQuery{
		Name:         &source.Spec.Resource.Name,
		AppNamespace: &source.Spec.Resource.Namespace,
		Revision:     ptr.To(source.Status.MainBranch),
	}
	before, err := getManifestsFromArgo(ctx, qMain, appCl)
	if err != nil {
		return reconcileResult, err
	}
	qNew := application.ApplicationManifestQuery{
		Name:         &source.Spec.Resource.Name,
		AppNamespace: &source.Spec.Resource.Namespace,
		Revision:     ptr.To(diff.Spec.TargetRef),
	}
	after, err := getManifestsFromArgo(ctx, qNew, appCl)
	if err != nil {
		return reconcileResult, err
	}

	dmp := diffmatchpatch.New()

	diffs := dmp.DiffMain(before, after, true)

	secretName := fmt.Sprintf("%s-diff", diff.Name)
	diffSecret := &corev1.Secret{
		ObjectMeta: metav1.ObjectMeta{
			Name:      secretName,
			Namespace: req.Namespace,
		},
	}

	err = r.Client.Get(ctx, types.NamespacedName{Namespace: req.Namespace, Name: secretName}, diffSecret)
	if err != nil {
		if !k8serrors.IsNotFound(err) {
			return reconcileResult, err
		}
		if err := r.Client.Create(ctx, diffSecret); err != nil {
			return reconcileResult, err
		}
		if err := r.Client.Get(ctx, types.NamespacedName{Namespace: req.Namespace, Name: secretName}, diffSecret); err != nil {
			return reconcileResult, err
		}
	}
	diffSecret.Data = map[string][]byte{}
	diffSecret.Data["diff"] = []byte(dmp.DiffPrettyText(diffs))
	if err := r.Update(ctx, diffSecret); err != nil {
		return reconcileResult, err
	}

	return reconcileResult, nil
}

func getManifestsFromArgo(ctx context.Context, q application.ApplicationManifestQuery, appCl application.ApplicationServiceClient) (string, error) {
	log := log.FromContext(ctx)
	log.Info("requesting an app", "application", q.Name, "appNamespace", q.AppNamespace)
	var resources string

	res, err := appCl.GetManifests(ctx, &q)
	if err != nil {
		return "", err
	}
	var unstructureds []*unstructured.Unstructured
	for _, mfst := range res.Manifests {
		obj, err := argoappv1.UnmarshalToUnstructured(mfst)
		if err != nil {
			return "", err
		}
		unstructureds = append(unstructureds, obj)
	}

	for _, rawManifest := range unstructureds {
		if rawManifest.Object != nil {
			yamlBytes, err := yaml.Marshal(rawManifest.Object)
			if err != nil {
				return "", err
			}

			resources += fmt.Sprintf("---\n%s\n", string(yamlBytes))

			app := &argoappv1.Application{}
			if err := yaml.Unmarshal(yamlBytes, app); err != nil {
				log.Info(err.Error())
			} else {
				log.Info(fmt.Sprintf("%v", app))
				app.ObjectMeta.Name = fmt.Sprintf("%s-%s", "diff", app.ObjectMeta.Name)
				app.Spec.SyncPolicy = &argoappv1.SyncPolicy{}

				cl, err := appCl.GetManifestsWithFiles(ctx)
				if err != nil {
					return "", err
				}

				chunk := &application.ApplicationManifestQueryWithFilesWrapper_Chunk{
					Chunk: &application.FileChunk{
						Chunk: yamlBytes,
					},
				}
				if err := cl.Send(&application.ApplicationManifestQueryWithFilesWrapper{
					Part: chunk,
				}); err != nil {
					return "", err
				}
				rawManifest, err := cl.CloseAndRecv()
				if err != nil {
					return "", err
				}
				for _, m := range rawManifest.Manifests {
					resources += fmt.Sprintf("---\n%s\n", m)
				}
			}
		}
	}
	return resources, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *DiffReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&k8sbadhouseplantsnetv1alpha1.Diff{}).
		Complete(r)
}
